import { ipcMain, IpcMainInvokeEvent } from 'electron';
import path from 'path';
import os from 'os';
import {spawn, ChildProcess} from 'child_process';
import {BASE_WIDTH, BASE_HEIGHT, BASE_LONG_SIDE, BASE_MOVEMENT_AMOUNT} from './settings';

type Padding = {
  verticalPad: number;
  horizontalPad: number;
}

function getSobelMovement(width: number, baseLongSide: number, baseMovementAmount: number): number{
  return Math.round((width / baseLongSide) * baseMovementAmount);
}

function calcPadding(finalWidth: number, height: number, baseWidth: number, baseHeight: number): Padding{
  if(finalWidth / height > baseWidth / baseHeight){
    const ratedHeight = baseHeight * (finalWidth / baseWidth);
    const verticalPad = Math.round((ratedHeight - height) / 2);
    return {verticalPad, horizontalPad: 0};
  }else{
    const ratedWidth = baseWidth * (height / baseHeight)
    const horizontalPad = Math.round((ratedWidth - finalWidth) / 2)
    return {verticalPad: 0, horizontalPad};
  }
}

function getMillSecondsFromHMSM(hour: number, minute: number, second: number, millsecond: number): number{
  if(!isNaN(hour) && !isNaN(minute) && !isNaN(second) && !isNaN(millsecond)){
    return hour  * 3600 * 1000 + minute * 60 * 1000 + second * 1000 + millsecond; 
  }else{
    return -1
  }
}

function getDurationMsecFromLines(lines: Array<string>): number{
  const durations = (lines.filter((l) => {
    return !!l.match(/Duration: (\d+):(\d+):(\d+)\.(\d+)/);
  }));
  if(durations.length > 0){
    const duration = durations.pop();
    const match = duration.match(/Duration: (\d+):(\d+):(\d+)\.(\d+)/);
    const hour = parseInt(match[1]);
    const minute = parseInt(match[2]);
    const second = parseInt(match[3]);
    const millsecond = parseInt(match[4]);
    return getMillSecondsFromHMSM(hour, minute, second, millsecond);
  }else{
    return -1;
  }
}

function getLastTimeMsecFromLines(lines: Array<string>): number{
  const times = (lines.filter((l) => {
    return !!l.match(/time=(\d+):(\d+):(\d+)\.(\d+)/);
  }));
  if(times.length > 0){
    const time = times.pop();
    const match = time.match(/time=(\d+):(\d+):(\d+)\.(\d+)/);
    const hour = parseInt(match[1]);
    const minute = parseInt(match[2]);
    const second = parseInt(match[3]);
    const millsecond = parseInt(match[4]);
    return getMillSecondsFromHMSM(hour, minute, second, millsecond);
  }else{
    return -1;
  }
}

let ffmpegRef: ChildProcess = undefined;

async function ffmpegExec(event: IpcMainInvokeEvent, videoPathArray: Array<string>, destinationPathArray: Array<string>, dimensionArray: number[], cpuUsageLimit: string): Promise<number>{
  const threads = ((cpuUsageLimit): number=>{
    const numOfCpus = os.cpus().length
    if(cpuUsageLimit === 'FULL'){
      return numOfCpus;
    }else if(cpuUsageLimit === 'HALF'){
      return Math.max(1, Math.floor(numOfCpus / 2));
    }else{
      return 1;
    }
  })(cpuUsageLimit)
  let videoPath = path.join(...videoPathArray);
  if(os.platform() !== 'win32'){
    videoPath = '/' + videoPath
  }
  const videoPathFileName = path.parse(videoPath).base;
  let destinationPath = path.join(...destinationPathArray);
  if(os.platform() !== 'win32'){
    destinationPath = '/' + destinationPath
  }

  const rotation = dimensionArray[2];
  const flip = dimensionArray[3];
  const videoMode = dimensionArray[4] === 1;
  let hFlip = '';
  if(flip === -1){
    hFlip = 'hflip,';
  }
  const rotCount = rotation / 90;
  let transpose = '';
  if(!videoMode){
    for(let i=0;i<rotCount;i++){
      transpose += 'transpose=2,'
    }
  }
  let width = dimensionArray[0];
  let height = dimensionArray[1];
  if(rotCount % 2 === 1){
    width = dimensionArray[1];
    height = dimensionArray[0];
  }
  const sobelMovement = getSobelMovement(width, BASE_LONG_SIDE, BASE_MOVEMENT_AMOUNT);
  const finalWidth = width * 2;
  const { verticalPad, horizontalPad } = calcPadding(finalWidth, height, BASE_WIDTH, BASE_HEIGHT);

  const ffmpegPath = path.join(__dirname, 'bin', `ffmpeg-${os.platform()}`, 'ffmpeg');

  const ffmpegParam = [
    '-y',
    '-i',
    videoPath,
    '-threads',
    `${threads}`,
    '-filter_complex',
    [
      `[0] setsar=1/1,${hFlip}${transpose}scale=w=${width}:h=${height},format=rgba,split [main_src][sobel_src];`,
      `[sobel_src] sobel=15:1:0 [sobel_green];`,
      `[sobel_green] format=gray,gblur=sigma=8:sigmaV=2,format=rgba[sobl_gray_src];`,
      `[sobl_gray_src] split [sobl_gray_merge_src][sobl_gray_inv_src];`,
      `[sobl_gray_inv_src] negate [sobl_gray_inv];`,
      `[sobl_gray_inv][sobl_gray_merge_src] alphamerge [sobl_gray];`,
      `[sobl_gray] split [sobel_left][sobel_right];`,
      `[main_src] split [left][right];`,
      `[left] pad=width=iw+${horizontalPad*2}:x=${horizontalPad}:height=ih+${verticalPad*2}:y=${verticalPad}:color=black [left_padded];`,
      `[right] pad=width=iw+${horizontalPad*2}:x=${horizontalPad}:height=ih+${verticalPad*2}:y=${verticalPad}:color=black [right_padded];`,
      `[left_padded][sobel_left] overlay=x=${horizontalPad-sobelMovement}:y=${verticalPad} [sobel_left_comp];`,
      `[right_padded][sobel_right] overlay=x=${horizontalPad+sobelMovement}:y=${verticalPad} [sobel_right_comp];`,
      `[sobel_left_comp] pad=width=iw*2:color=black [main_with_left];`,
      `[main_with_left][sobel_right_comp] overlay=x=${width+horizontalPad*2} [result];`,
      `[result] format=yuv420p`,
    ].join(''),
    path.join(destinationPath, `vr-${videoPathFileName}`)
  ]

  const ffmpeg = spawn(ffmpegPath,ffmpegParam);

  ffmpegRef = ffmpeg;

  let stderr = '';

  return new Promise<number>((resolve) => {
    ffmpeg.stderr.on("data", data => {
      stderr += data.toString();
      console.error(`${data}`);
      // ffmpeg send CR and LF for some reason because replace to LF all.
      // ffmpegがなぜかCRとLFを使い分けて送ってくるので全部LFに変える。
      const lines = stderr.replace(/\r/g, '\n').split('\n');
      const duration = getDurationMsecFromLines(lines);
      const time = getLastTimeMsecFromLines(lines);
      if(duration > 0 && time > 0){
        const completeRatio = (time / duration) * 100;
        event.sender.send('progress-update', completeRatio);
      }else{
        event.sender.send('progress-update', 0);
      }
    });

    ffmpeg.on("close", (code) => {
      ffmpegRef = undefined;
      resolve(code);
    });
  })
}

ipcMain.handle('ffmpeg-exec', async (event, videoPathArray: Array<string>, destinationPathArray: Array<string>, dimensionArray: number[], cpuUsageLimit: string): Promise<number> => {
  return await ffmpegExec(event, videoPathArray, destinationPathArray, dimensionArray, cpuUsageLimit);
});

ipcMain.handle('ffmpeg-kill', async (): Promise<boolean> => {
  return new Promise<boolean>((resolve) => {
    if(ffmpegRef){
      const refCopy = ffmpegRef;
      ffmpegRef = undefined;
      resolve(refCopy.kill())
    }else{
      resolve(false);
    }
  });
});
