import { ipcMain } from 'electron';
import path from 'path';
import os from 'os';
import {spawn, ChildProcess} from 'child_process';

// function getShortSide(longSide: number, shortSide: number, baseLongSide: number): number{
//   return Math.round(baseLongSide / longSide * shortSide);
// }

// function calcMediaDimensions(width: number, height: number): Array<number>{
//   if((width > BASE_LONG_SIDE) || (height > BASE_LONG_SIDE)){
//     if(width >= height){
//       height = getShortSide(width, height, BASE_LONG_SIDE);
//       width = BASE_LONG_SIDE;
//     }else{
//       height = BASE_LONG_SIDE;
//       width = getShortSide(height, width, BASE_LONG_SIDE);
//     }
//   }
//   return [width, height];
// }

let ffprobeRef: ChildProcess = undefined
let ffprobeExecutable = true;

async function ffprobeExec(videoPathArray: Array<string>, commandArguments: Array<string>, processStdout: (stdout: string) => number[] ): Promise<number[]>{
  let videoPath = path.join(...videoPathArray);
  if(os.platform() !== 'win32'){
    videoPath = '/' + videoPath
  }

  const ffprobePath = path.join(__dirname, 'bin', `ffmpeg-${os.platform()}`, 'ffprobe');

  const ffprobe = spawn(ffprobePath, commandArguments.concat([videoPath]));

  ffprobeRef = ffprobe;

  return new Promise((resolve) => {
    let stdout = '';
    ffprobe.stdout.on("data", data => {
      stdout += data;
    });
  
    ffprobe.on("close", () => {
      const result = processStdout(stdout);
      ffprobeRef = undefined;
      resolve(result)
    });
  })
}

async function ffprobeExecFetchDimension(videoPathArray: Array<string>): Promise<number[]>{
  return await ffprobeExec(videoPathArray, ['-v', 'error', '-show_entries', 'stream=width,height', '-of', 'default=noprint_wrappers=1'], (stdout) => {
    const lines = stdout.replace("\r", '').split('\n');
    let width = 0;
    let height = 0;
    lines.forEach((l) => {
      if(l.startsWith('width=')){
        width = parseInt(l.split('=')[1])
      }else if(l.startsWith('height=')){
        height = parseInt(l.split('=')[1])
      }
    })
    return [width, height];
  });
}

async function ffprobeExecFetchRotationVideo(videoPathArray: Array<string>): Promise<number[]>{
  // ref: https://superuser.com/questions/660505/find-out-rotation-metadata-from-video-in-ffmpeg
  return await ffprobeExec(videoPathArray, ['-v', 'error', '-show_entries', 'stream_tags=rotate', '-of', 'default=nw=1:nk=1'], (stdout) => {
    const lines = stdout.replace("\r", '').split('\n');
    let rotationTowardCCW = 0;
    lines.forEach((l) => {
      const trimSpace = l.replace(/\s/g, '');
      if(trimSpace.length > 0){
        const parsedRotation = parseInt(trimSpace);
        if(!isNaN(parsedRotation)){
          rotationTowardCCW = parsedRotation;
        }
      }
    })
    return [rotationTowardCCW, 1, 1];
  });
}

function convertRotationFlipFromNumber(rotationCode: number): number[]{
  // ref: https://jdhao.github.io/2019/07/31/image_rotation_exif_info/
  switch(rotationCode){
    case 1:
      return [0, 1, 0]
    case 2:
      return [0, -1, 0]
    case 3:
      return [180, 1, 0]
    case 4:
      return [180, -1, 0]
    case 5:
      return [90, -1, 0]
    case 6:
      return [270, 1, 0]
    case 7:
      return [270, -1, 0]
    case 8:
      return [90, 1, 0]
  }
  return [0, 1, 0]
}

async function ffprobeExecFetchRotationImage(videoPathArray: Array<string>): Promise<number[]>{
  // ref: https://www.exiftool.org/TagNames/EXIF.html
  return await ffprobeExec(videoPathArray, ['-v', 'error', '-show_entries', 'frame_tags=Orientation', '-of', 'default=nw=1:nk=1'], (stdout) => {
    const lines = stdout.replace("\r", '').split('\n');
    let rotationCode = 0;
    lines.forEach((l) => {
      const trimSpace = l.replace(/\s/g, '');
      if(trimSpace.length > 0){
        const parsedRotation = parseInt(trimSpace);
        if(!isNaN(parsedRotation)){
          rotationCode = parsedRotation;
        }
      }
    })
    return convertRotationFlipFromNumber(rotationCode);
  });
}

ipcMain.handle('ffprobe-exec', async (_, videoPathArray: Array<string>): Promise<number[]> => {
  const dimension =  await ffprobeExecFetchDimension(videoPathArray);
  if(!ffprobeExecutable){
    return dimension.concat([0, 1]);
  }
  if(videoPathArray[videoPathArray.length-1].match(/(\.mp4|\.mov)$/)){
    const rotation = await ffprobeExecFetchRotationVideo(videoPathArray);
    return dimension.concat(rotation);
  }else{
    const rotation = await ffprobeExecFetchRotationImage(videoPathArray);
    return dimension.concat(rotation);
  }
});

ipcMain.handle('ffprobe-kill', async (): Promise<boolean> => {
  return new Promise<boolean>((resolve) => {
    if(ffprobeRef){
      ffprobeExecutable = false;
      const refCopy = ffprobeRef;
      ffprobeRef = undefined;
      const result = refCopy.kill();
      ffprobeExecutable = true;
      resolve(result)
    }else{
      resolve(false);
    }
  });
});
